unit complex_sub;

interface

uses Velthuis.BigIntegers, Velthuis.Bigdecimals;

type
  combig = record
    r : bigdecimal;
    i : bigdecimal;
  end;

const
  EPSC = '1e-150';
  DPS = 200;

var
  zero : bigdecimal;     // 0
  one, two : bigdecimal;      // 1
  precisions : integer = DPS;        // L


function combiground(x : combig): combig;
function cadd(a, b: comBig): ComBig;
function csub(a, b: ComBig): ComBig;
function caddb(a: combig; b:bigdecimal): combig;
function csubb(a : combig; b: bigdecimal): combig;
function cmulb(a: Combig; b: bigdecimal): ComBig;
function cmul(a, b: Combig): ComBig;
function cdivb(a: Combig; b: bigdecimal): Combig;
function cdiv(a, b: Combig): Combig;
function caeqb(x, y : combig): boolean;
function cabs(a : combig): bigdecimal;
function pow_big(x, y : combig; eps : bigdecimal): combig;
function pi_big: Bigdecimal;
function log_big(z : combig; eps : bigdecimal; dpcs: integer): combig;
function exp_big(x: Combig; dpcs: integer; eps : bigdecimal): Combig;
function cchs(x: combig): combig;
function cdivi(a: Combig; b: biginteger): Combig;


implementation


//------------------------------  comprex calc-------------------------------------------
var
  z10 : combig;
  Z10F : boolean = false;

// x: combig dpcsɊۂ
function combiground(x : combig): combig;
begin
  result.r := x.r.RoundToPrecision(precisions);
  result.i := x.i.RoundToPrecision(precisions);
end;

// ComBigf̉Z
function cadd(a, b: comBig): ComBig;
begin
  result.r := a.r + b.r;
  result.i := a.i + b.i;
end;

// ComBigf̌Z
function csub(a, b: ComBig): ComBig;
begin
  result.r := a.r - b.r;
  result.i := a.i - b.i;
end;

// combig + bigdecimal
function caddb(a: combig; b:bigdecimal): combig;
begin
  result.r := a.r + b;
  result.i := a.i;
end;

// combig - bigdecimal
function csubb(a : combig; b: bigdecimal): combig;
begin
  result.r := a.r - b;
  result.i := a.i;
end;

// Combig bigdecimal̏Z
function cmulb(a: Combig; b: bigdecimal): ComBig;
begin
  result.r := a.r * b;
  result.i := a.i * b;
end;

// Combigf̏Z
function cmul(a, b: Combig): ComBig;
begin
  result.r := a.r * b.r - a.i * b.i;
  result.i := a.r * b.i + a.i * b.r;
  result := combiground(result);
end;

// Combig bigdecimal̏Z
function cdivb(a: Combig; b: bigdecimal): Combig;
begin
  if b <> Bigdecimal.Zero then begin
    result.r := a.r / b;
    result.i := a.i / b;
  end
  else begin
    result.r := Bigdecimal.Zero;
    result.i := Bigdecimal.Zero;
  end;
end;

function cdivi(a: Combig; b: biginteger): Combig;
begin
  if b <> biginteger.Zero then begin
    result.r := a.r / b;
    result.i := a.i / b;
  end
  else begin
    result.r := Bigdecimal.Zero;
    result.i := Bigdecimal.Zero;
  end;
end;

// Combigf̏Z
function cdiv(a, b: Combig): Combig;
var
  bb, arbraibi, aibrarbi: Bigdecimal;
begin
  bb := b.r * b.r + b.i * b.i;
  arbraibi := a.r * b.r + a.i * b.i;
  aibrarbi := a.i * b.r - a.r * b.i;
  bb := bb.RoundToPrecision(precisions);
  arbraibi := arbraibi.RoundToPrecision(precisions);
  aibrarbi := aibrarbi.RoundToPrecision(precisions);

  if bb <> Bigdecimal.Zero then begin
    result.r := arbraibi / bb;
    result.i := aibrarbi / bb;
  end
  else begin
    result.r := Bigdecimal.Zero;
    result.i := Bigdecimal.Zero;
  end;
end;

// a:combig = b:combig
function caeqb(x, y : combig): boolean;
begin
  if (x.r = y.r) and (x.i = y.i) then result := true
                                 else result := false;
end;

// Combig̐Βl
function cabs(a : combig): bigdecimal;
var
  arh2, aih2 : bigdecimal;
  tmp : bigdecimal;
begin
  arh2 := a.r * a.r;
  aih2 := a.i * a.i;
  arh2 := arh2.RoundToPrecision(precisions);
  aih2 := aih2.RoundToPrecision(precisions);
  tmp := arh2 + aih2;
  result := bigdecimal.Sqrt(tmp, precisions);
end;

//------------------------------------------------------------------------------
// 
// https://www.kurims.kyoto-u.ac.jp/~ooura/pi_fft-j.html
// T.Ooura AGM ASY
function pi_big: Bigdecimal;
var
  n, dpcs : integer;
  rmback : BigDecimal.RoundingMode;
  SQRT_SQRT_EPSILON, c, b, e : BigDecimal;
  npow : Bigdecimal;
  a, one, two, four, five, eight : Bigdecimal;
begin
  rmBack := BigDecimal.DefaultRoundingMode;           // Zۂ߃[hobNAbv
  dpcs := precisions + 5;                             // w萸x + 
  BigDecimal.DefaultPrecision := dpcs;
  BigDecimal.DefaultRoundingMode := rmNearesteven;

  one := '1';
  one := one.RoundToPrecision(dpcs);                  // one̗Ldpcsɐݒ
  two := '2';
  four := '4';
  five := '5';
  eight := '8';
  SQRT_SQRT_EPSILON := one / BigDecimal.IntPower(two, dpcs shl 0); // l
  c := BigDecimal.Sqrt(one / eight, dpcs * 2);
  a := one + c + c + c;
  b := BigDecimal.Sqrt(a, dpcs * 2);
  e := b - five / eight;
  b := b + b;
  c := e - c;
  a := a + e;
  npow := '4';
  n := 0;
  while (e > SQRT_SQRT_EPSILON) and (n < 100) do begin
    npow := npow + npow;                              // 8,16,32,64
    e := (a + b) / two;                               // ϒl ZLł̊ۂ
    b := a * b;
    b := b.RoundToPrecision(dpcs);                    // pcs +  ۂ
    b := BigDecimal.Sqrt(b, dpcs * 2);                // Lł̊ۂ
    e := e - b;
    b := b + b;
    c := c - e;
    a := e + b;
    inc(n);
  end;
  e := e * e;
  e := e.RoundToPrecision(dpcs);                      // pcs +  ۂ
  e := e / four;                                      // e = e * e / 4
  a := a + b;
  result := (a * a - e - e / two) / (a * c - e) / npow;   // Z̏Ԃɒ
  result := result.RoundToPrecision(precisions);             // w̐xɊۂ

  BigDecimal.DefaultPrecision := precisions;                 // ZZxA
  BigDecimal.DefaultRoundingMode := rmBack;           // Zۂ߃[hA
end;

// z ͒l f
// eps l
// dpcs L
// ndis \ true@\ false \
function log_big_sub(z : Combig; eps : bigdecimal; dpcs: integer; ndis: boolean): Combig;
const
  NN = 1000000;
var
  s, x, zm1, zp1, xh2, xg, n2p1 : ComBig;
  one, two : Combig;
  xgsn : Combig;
  asabs : bigdecimal;
  n : integer;
begin
  one.r := Bigdecimal.One;      // 1
  one.i := Bigdecimal.Zero;     //
  two := cadd(one, one);        // 2
  zm1 := csub(z, one);          // z - 1
  zp1 := cadd(z, one);          // z + 1
  x := cdiv(zm1, zp1);          // x = (z-1)/(z+1)
  s.r := Bigdecimal.Zero;
  s.i := Bigdecimal.Zero;
  xh2 := cmul(x, x);            // x^2
  xg := x;                      // n = 0;    x^(2n+1)
  n2p1 := one;                  // 2n+1 = 1
  n := 0;
  repeat
    xg.r := xg.r.RoundToPrecision(dpcs);
    xg.i := xg.i.RoundToPrecision(dpcs);
    xgsn := cdiv(xg, n2p1);     // x^(2n+1) / (2n+1)
    s := cadd(s, xgsn);         // 
    xg := cmul(xg, xh2);        // x^(2n+1)
    n2p1 := cadd(n2p1, two);    // 2n + 1
    inc(n);
    asabs := cabs(xgsn);
  until  (asabs < eps) or (n > NN);
  s := cmul(s, two);
  result := s;
end;

// z   
// eps l
// dpcs L
// Ԃl ΐf
function log_big(z : combig; eps : bigdecimal; dpcs: integer): combig;
var
  zb : combig;
  z10n : combig;
  Zmax, ten, one, bn : bigdecimal;
  pi_big2: bigdecimal;
begin
  if (z.r = bigdecimal.Zero) and (z.i = bigdecimal.Zero) then begin
    result.r := bigdecimal.Zero;
    result.i := bigdecimal.Zero;
    exit;
  end;

  pi_big2 := pi_big / bigdecimal.Two;     // paib = pi_bigs  main.pas

  zb := z;

  // ƎƂ[łȂȂ 傫ȒlɂĉZ덷܂
  if (zb.r <> bigdecimal.Zero) and (zb.i <> bigdecimal.Zero) then begin
    if zb.i.Abs(zb.i) > zb.r.Abs(zb.r) then begin     // ̐Βl̐Βl傫ւ
      z.r := zb.i;
      z.i := zb.r;
      if zb.i < bigdecimal.Zero then begin            // 畄]
        z.r := -z.r;
        z.i := -z.i;
      end;
    end
    else begin                                        // ̐Βl̐ΒlƓ
      if zb.r < bigdecimal.Zero then begin            // ̒l畄]
        z.r := -zb.r;
        z.i := -zb.i;
      end;
    end;
  end;

  // z.r0z.i[łȂ  Ǝւ
  if (zb.r = bigdecimal.Zero) and (zb.i <> bigdecimal.Zero) then begin
    z.r := zb.i;
    z.i := zb.r;
    if z.r < bigdecimal.Zero then z.r := -z.r;      // 畄] [Ȃ̂ł̂܂
  end;

  // ̂ŋ[̎𔽓] [Ȃ̂ł̂܂
  if (zb.r <> bigdecimal.Zero) and (zb.i = bigdecimal.Zero) then
    if zb.r < bigdecimal.Zero then z.r := -zb.r;

  // 10̑ΐ𗘗pđ傫ȒľvZ𑬂܂
  ten := '10';
  ten := ten.RoundToPrecision(dpcs);               // dpcs킹
  one := bigdecimal.One;
  bn := bigdecimal.Zero;
  if not Z10F then begin
    z10.r := ten;                                       // z10 : cbig Z!0F : boolean  main.pas
    z10.i := bigdecimal.Zero;
    z10 := log_big_sub(z10, eps, dpcs, false);          // 10̑ΐ
    Z10F := true;
  end;
  zmax := cabs(z);                                // z̐Βl
  zmax := zmax.RoundToPrecision(dpcs);             // dpcs킹
  z.r := z.r.RoundToPrecision(dpcs);               // dpcs킹
  z.i := z.i.RoundToPrecision(dpcs);               // dpcs킹
  // 10ŏZ͏ZĒl̒l1`10͈̔͂ɓ悤ɂ܂B
  repeat
    if zmax > ten then begin        // zmax10傫10ŏZ
      z.r := z.r / ten;
      z.i := z.i / ten;
      zmax := zmax / ten;
      bn := bn + one;               // 10ŏZ +ɂȂ܂B
    end;
    if zmax < one then begin        // zmax1菬10Z
      z.r := z.r * ten;
      z.i := z.i * ten;
      zmax := zmax * ten;
      bn := bn - one;               // 10Z -ɂȂ܂B
    end;
  until (zmax <= ten) and (zmax >= one);
  //----------------------------------------

  // CꂽlőΐvZ
  result := log_big_sub(z, eps, dpcs, true);

  // 10̑ΐ␳
  z10n.r := z10.r * bn;             // ZΐlvZ10̑ΐln{
  z10n.i := z10.i * bn;

  result.r := result.r + z10n.r;          // 10̑ΐln{Z
  result.i := result.i + z10n.i;
  //----------------------------------------

  // z.r[z.i[łȂ /2␳
  if (zb.r = bigdecimal.Zero) and (zb.i <> bigdecimal.Zero) then begin
    if zb.i >= bigdecimal.Zero then result.i := result.i + pi_big2
                               else result.i := result.i - pi_big2;
  end;

  // z.r̒l̂ƕύXČvZĂ̋Ε␳B
  if (zb.r < bigdecimal.Zero) and (zb.i = bigdecimal.Zero) then
    result.i := result.i + pi_big;

  // ƎƂ[łȂ
  if (zb.r <> bigdecimal.Zero) and (zb.i <> bigdecimal.Zero) then begin
    // Βl苕̐Βl傫ꍇ
    if zb.i.Abs(zb.i) > zb.r.Abs(zb.r) then begin
      if zb.i > bigdecimal.Zero then result.i := -result.i + pi_big2    // /2␳
                                else result.i := -result.i - pi_big2;
    end
    // ̎
    else begin
      if zb.r < bigdecimal.Zero then
        if zb.i > bigdecimal.Zero then result.i := result.i + pi_big    // Ε␳
                                  else result.i := result.i - pi_big
    end;
  end;
end;


// exp(x)    eC[
// x ͒l f
// dpcs L
// eps l
// x / 8 ->  ans^8    8ŏZ vZ Ō8
function exp_big(x: Combig; dpcs: integer; eps : bigdecimal): Combig;
var
  xh, s, si: Combig;
  one, ass, eight, i, ih : bigdecimal;
begin
  one := bigdecimal.One;
  eight := bigdecimal.Ten - bigdecimal.Two;
  eight := eight.RoundToPrecision(dpcs);
  x.r := x.r.RoundToPrecision(dpcs);
  x.i := x.i.RoundToPrecision(dpcs);
  x := cdivb(x, eight);      // x/8
  i := one;
  ih := i;
  xh := x;
  s := x;
  repeat
    i := i + one;
    ih := ih * i;             // i!
    xh := cmul(xh, x);        // x^n
    xh.r := xh.r.RoundToPrecision(dpcs);
    xh.i := xh.i.RoundToPrecision(dpcs);
    si := cdivb(xh, ih);      // x^n/i!
    s := cadd(s, si);
    ass := cabs(si);
  until (ass < eps);
  s.r := s.r + one;
  s := cmul(s, s);            // x^2
  s := cmul(s, s);            // x^4
  result := cmul(s, s);       // x^8
  result.r := result.r.RoundToPrecision(dpcs);
  result.i := result.i.RoundToPrecision(dpcs);
end;

// x^y
// exp[ln(x) * y]
// A[ɂȂłAZ덷ɂA[ɂȂȂ̂āAċI
// Ƀ[ɂĂ܂B
function pow_big(x, y : combig; eps : bigdecimal): combig;
var
  dpcs : integer;
  harf, absi, absy, four : bigdecimal;
  lnx, tmp : combig;
begin
  if (x.r = bigdecimal.zero) and (x.i = bigdecimal.zero) then begin
    result.r := bigdecimal.Zero;
    result.i := bigdecimal.Zero;
    exit;
  end;
  dpcs := BigDecimal.DefaultPrecision;
  lnx := log_big(x, eps, dpcs);               // ln(x)
  lnx := combiground(lnx);                // ppcs ۂ
  y := combiground(y);                    // ppcs ۂ
  tmp := cmul(lnx, y);
  tmp := combiground(tmp);                // ln(x) * y
  result := exp_big(tmp, dpcs, eps);

  // x̋y̋0Ȃ
  if (x.i = bigdecimal.zero) and (y.i = bigdecimal.Zero) then begin
    harf := bigdecimal.One / bigdecimal.Two;          // 0.5
    absy := y.r.Abs(y.r);       // y̎̐Βl
    absi := absy.Frac;          // y̎̐Βl̏
    absi := absi - harf;        // y̎̐Βl̏0.5̍
    // [(y̎̏0.5)
    if absi = bigdecimal.Zero then
      // x̎0Ɠ0傫Ȃ
      if x.r >= bigdecimal.zero then result.i := bigdecimal.zero   // ̋0
                                // 0菬Ȃ
                                else result.r := bigdecimal.zero;  // ̎0
    absi := y.r.Frac;           // 搔̏
    if absi = bigdecimal.Zero then result.i := bigdecimal.Zero;    // Ȃ瓚̋0
  end;
  // x̐y̋0Ȃ
  if (x.r = bigdecimal.Zero) and (y.i = bigdecimal.Zero) then begin
    absi := y.r.Frac;                                     // 搔̏
    if absi = bigdecimal.Zero then begin                  // Ȃ
      absy := y.r / bigdecimal.Two;                       // mF
      absi := absy.Frac;                                  // 
      if absi = bigdecimal.Zero then result.i := bigdecimal.Zero             // Ȃ狕0
                                else result.r := bigdecimal.Zero;            // Ȃ0
    end;
  end;
  absy := x.r.Abs(x.r);                                     // |x.r|
  absi := x.i.Abs(x.i);                                     // |x.i|
  // x̎Ƌ̐Βly.r[łȂ
  if (absy = absi) and (y.i = bigdecimal.Zero) and not (y.r = bigdecimal.Zero) then begin
    absi := y.r.Frac;                                       // 搔̏
    if absi = bigdecimal.Zero then  begin                   // Ȃ
      absy := y.r / bigdecimal.Two;                         // mF
      absi := absy.Frac;                                    // 
      if absi = bigdecimal.Zero then begin                  // Ȃ
        four := bigdecimal.Two + bigdecimal.Two;            // 4
        absy := y.r / four;                                 // 4mF
        absi := absy.Frac;                                  // 
        if absi = bigdecimal.Zero then result.i := bigdecimal.Zero      // Ȃ 0  y= 4 8
                                  else result.r := bigdecimal.Zero;     // Ȃ 0  y= 2 6 10
      end;
    end;
  end;
end;

// ]
function cchs(x: combig): combig;
begin
  result.r := -x.r;
  result.i := -x.i;
end;

initialization
  zero := bigdecimal.Zero;
  one := bigdecimal.One;
  two := bigdecimal.Two;
end.
